local VERSION = tonumber(("$Revision: 44393 $"):match("%d+"))

local Parrot = Parrot
local Parrot_Display = Parrot:NewModule("Display", "AceEvent-2.0", "AceHook-2.1")
if Parrot.revision < VERSION then
	Parrot.version = "r" .. VERSION
	Parrot.revision = VERSION
	Parrot.date = ("$Date: 2007-07-20 22:14:36 -0400 (Fri, 20 Jul 2007) $"):match("%d%d%d%d%-%d%d%-%d%d")
end

local L = Parrot:L("Parrot_Display")

local SharedMedia = AceLibrary("SharedMedia-1.0")

local new, del = Parrot.new, Parrot.del

local ParrotFrame

Parrot_Display.db = Parrot:AcquireDBNamespace("Display")
Parrot:RegisterDefaults("Display", 'profile', {
	alpha = 1,
	font = "Friz Quadrata TT",
	fontSize = 18,
	fontOutline = "THICKOUTLINE",
	stickyFont = "Friz Quadrata TT",
	stickyFontSize = 26,
	stickyFontOutline = "THICKOUTLINE",
})

local Parrot_AnimationStyles
local Parrot_Suppressions
local Parrot_ScrollAreas
function Parrot_Display:OnEnable()
	Parrot_AnimationStyles = Parrot:GetModule("AnimationStyles")
	Parrot_Suppressions = Parrot:GetModule("Suppressions")
	Parrot_ScrollAreas = Parrot:GetModule("ScrollAreas")
	if not ParrotFrame then
		ParrotFrame = CreateFrame("Frame", "ParrotFrame", UIParent)
		ParrotFrame:SetFrameStrata("HIGH")
		ParrotFrame:SetToplevel(true)
		ParrotFrame:SetPoint("CENTER")
		ParrotFrame:SetWidth(0.0001)
		ParrotFrame:SetHeight(0.0001)
	end
	
	if _G.CombatText_AddMessage then
		self:Hook("CombatText_AddMessage", true)
	else
		function _G.CombatText_AddMessage(...)
			self:CombatText_AddMessage(...)
		end
	end
end

function Parrot_Display:OnDisable()
end

-- #NODOC
function Parrot_Display:CombatText_AddMessage(message, scrollFunction, r, g, b, displayType, isStaggered)
	if type(message) ~= "string" then
		return
	end
	if type(r) ~= "number" or type(g) ~= "number" or type(b) ~= "number" then
		r, g, b = 1, 1, 1
	end
	self:ShowMessage(message, "Notification", displayType == "crit", r, g, b, nil, nil, nil)
end

function Parrot_Display:OnOptionsCreate()
	local outlineValidate = {
		NONE = L["None"],
		OUTLINE = L["Thin"],
		THICKOUTLINE = L["Thick"],
	}
	
	Parrot.options.args.general.args.alpha = {
		type = 'range',
		name = L["Text transparency"],
		desc = L["How opaque/transparent the text should be."],
		min = 0.25,
		max = 1,
		step = 0.01,
		bigStep = 0.05,
		isPercent = true,
		get = function()
			return self.db.profile.alpha
		end,
		set = function(value)
			self.db.profile.alpha = value
		end
	}
	Parrot.options.args.general.args.font = {
		type = 'group',
		name = L["Master font settings"],
		desc = L["Master font settings"],
		args = {
			normalFace = {
				type = 'text',
				name = L["Normal font"],
				desc = L["Normal font face."],
				get = function()
					return Parrot_Display.db.profile.font
				end,
				set = function(value)
					Parrot_Display.db.profile.font = value
				end,
				validate = SharedMedia:List("font"),
			},
			normalSize = {
				type = 'range',
				name = L["Normal font size"],
				desc = L["Normal font size"],
				min = 12,
				max = 32,
				step = 1,
				get = function()
					return Parrot_Display.db.profile.fontSize
				end,
				set = function(value)
					Parrot_Display.db.profile.fontSize = value
				end,
			},
			normalBorder = {
				type = 'text',
				name = L["Normal outline"],
				desc = L["Normal outline"],
				validate = outlineValidate,
				get = function()
					return Parrot_Display.db.profile.fontOutline
				end,
				set = function(value)
					Parrot_Display.db.profile.fontOutline = value
				end,
			},
			stickyFace = {
				type = 'text',
				name = L["Sticky font"],
				desc = L["Sticky font face."],
				get = function()
					return Parrot_Display.db.profile.stickyFont
				end,
				set = function(value)
					Parrot_Display.db.profile.stickyFont = value
				end,
				validate = SharedMedia:List("font"),
			},
			stickySize = {
				type = 'range',
				name = L["Sticky font size"],
				desc = L["Sticky font size"],
				min = 12,
				max = 32,
				step = 1,
				get = function()
					return Parrot_Display.db.profile.stickyFontSize
				end,
				set = function(value)
					Parrot_Display.db.profile.stickyFontSize = value
				end,
			},
			stickyBorder = {
				type = 'text',
				name = L["Sticky outline"],
				desc = L["Sticky outline"],
				validate = outlineValidate,
				get = function()
					return Parrot_Display.db.profile.stickyFontOutline
				end,
				set = function(value)
					Parrot_Display.db.profile.stickyFontOutline = value
				end,
			},
		}
	}
end

local freeFontStrings = {}
local wildFontStrings
local fontString_num = 0
local fontStringIDs = {}

--[[----------------------------------------------------------------------------------
Arguments:
	string - the text you wish to show.
	[optional] string - the scroll area to show in. Default: "Notification"
	[optional] boolean - whether to show in the sticky-style, e.g. crits. Default: false
	[optional] number - [0, 1] the red part of the color. Default: 1
	[optional] number - [0, 1] the green part of the color. Default: 1
	[optional] number - [0, 1] the blue part of the color. Default: 1
	[optional] string - the font to use (as determined by SharedMedia-1.0). Defaults to the scroll area's setting.
	[optional] number - the font size to use. Defaults to the scroll area's setting.
	[optional] string - the font outline to use. Defaults to the scroll area's setting.
Notes:
	* See :GetScrollAreasValidate() for a validation list of scroll areas.
	* Messages are suppressed if the user has set a specific suppression matching the text.
Example:
	Parrot:ShowMessage("Hello, world!", "Notification", false, 0.5, 0.5, 1)
------------------------------------------------------------------------------------]]
function Parrot_Display:ShowMessage(text, scrollArea, sticky, r, g, b, font, fontSize, outline)
	self = Parrot_Display -- for those who do Parrot:ShowMessage
	if not Parrot:IsModuleActive(self) then
		return
	end
	if Parrot_Suppressions:ShouldSuppress(text) then
		return
	end
	if not Parrot_ScrollAreas:HasScrollArea(scrollArea) then
		if Parrot_ScrollAreas:HasScrollArea("Notification") then
			scrollArea = "Notification"
		else
			scrollArea = Parrot_ScrollAreas:GetRandomScrollArea()
			if not scrollArea then
				return
			end
		end
	end
	local fs = next(freeFontStrings)
	if fs then
		fs:ClearAllPoints()
		freeFontStrings[fs] = nil
	else
		fontString_num = fontString_num + 1
		fs = ParrotFrame:CreateFontString("ParrotFrameFontString" .. fontString_num, "ARTWORK", "MasterFont")
	end
	
	scrollArea = Parrot_ScrollAreas:GetScrollArea(scrollArea)
	if not sticky then
		if not font then
			font = scrollArea.font or self.db.profile.font
		end
		if not fontSize then
			fontSize = scrollArea.fontSize or self.db.profile.fontSize
		end
		if not outline then
			outline = scrollArea.fontOutline or self.db.profile.fontOutline
		end
	else
		if not font then
			font = scrollArea.stickyFont or self.db.profile.stickyFont
		end
		if not fontSize then
			fontSize = scrollArea.stickyFontSize or self.db.profile.stickyFontSize
		end
		if not outline then
			outline = scrollArea.stickyFontOutline or self.db.profile.stickyFontOutline
		end
	end
	if outline == "NONE" then
		outline = ""
	end
	fs:SetFont(SharedMedia:Fetch('font', font), fontSize, outline)
	if r and g and b then
		fs:SetTextColor(r, g, b)
	else
		fs:SetTextColor(1, 1, 1)
	end
	fs:SetText(text)
	fs.start = GetTime()
	fs.scrollArea = scrollArea
	fs.sticky = sticky
	local animationStyle
	if sticky then
		animationStyle = scrollArea.stickyAnimationStyle
	else
		animationStyle = scrollArea.animationStyle
	end
	local aniStyle = Parrot_AnimationStyles:GetAnimationStyle(animationStyle)
	fs:SetPoint("BOTTOM", 0, 0)
	if not wildFontStrings then
		wildFontStrings = new()
		self:ScheduleRepeatingEvent("Parrot_OnUpdate", self.OnUpdate, 0, self)
	end
	local wildFontStrings_scrollArea = wildFontStrings[scrollArea]
	if not wildFontStrings_scrollArea then
		wildFontStrings_scrollArea = new()
		wildFontStrings[scrollArea] = wildFontStrings_scrollArea
	end
	local wildFontStrings_scrollArea_aniStyle = wildFontStrings_scrollArea[aniStyle]
	if not wildFontStrings_scrollArea_aniStyle then
		wildFontStrings_scrollArea_aniStyle = new()
		wildFontStrings_scrollArea[aniStyle] = wildFontStrings_scrollArea_aniStyle
	end	
	wildFontStrings_scrollArea_aniStyle.length = scrollArea[sticky and "stickySpeed" or "speed"] or 3
	local fontStringIDs_scrollArea = fontStringIDs[scrollArea]
	if not fontStringIDs_scrollArea then
		fontStringIDs_scrollArea = new()
		fontStringIDs[scrollArea] = fontStringIDs_scrollArea
	end
	local fontStringIDs_scrollArea_aniStyle = fontStringIDs_scrollArea[aniStyle]
	if not fontStringIDs_scrollArea_aniStyle then
		fontStringIDs_scrollArea_aniStyle = 1
	else
		fontStringIDs_scrollArea_aniStyle = fontStringIDs_scrollArea_aniStyle + 1
	end
	fontStringIDs_scrollArea[aniStyle] = fontStringIDs_scrollArea_aniStyle
	fs.id = fontStringIDs_scrollArea_aniStyle
	
	table.insert(wildFontStrings_scrollArea_aniStyle, 1, fs)
	fs:Show()
	local init = aniStyle.init
	if init then
		init(fs, scrollArea, scrollArea[sticky and "stickyReverse" or "reverse"], fontStringIDs_scrollArea_aniStyle)
	end
	fs:SetAlpha(self.db.profile.alpha)
	self:OnUpdate(scrollArea, aniStyle)
end
Parrot.ShowMessage = Parrot_Display.ShowMessage

local function isOverlapping(alpha, bravo)
	if alpha:GetLeft() <= bravo:GetRight() and bravo:GetLeft() <= alpha:GetRight() and alpha:GetBottom() <= bravo:GetTop() and bravo:GetBottom() <= alpha:GetTop() then
		return true
	end
end

function Parrot_Display:OnUpdate()
	local now = GetTime()
	for scrollArea, u in pairs(wildFontStrings) do
		for animationStyle, t in pairs(u) do
			local t_len = #t
			local lastFS = nil
			for i, fs in ipairs(t) do
				local start, length = fs.start, t.length
				if start + length <= now then
					for j = i, t_len do
						local fs = t[j]
						local cleanup = animationStyle.cleanup
						if cleanup then
							cleanup(fs, scrollArea, scrollArea[fs.sticky and "stickyReverse" or "reverse"], fs.id)
						end
						fs:Hide()
						t[j] = nil
						freeFontStrings[fs] = true
					end
					break
				end
				local percent = (now - start) / length
				if percent >= 0.8 then
					fs:SetAlpha((1-percent) * 5 * self.db.profile.alpha)
				end
				
				fs:ClearAllPoints()
				animationStyle.func(fs, scrollArea, percent, scrollArea[fs.sticky and "stickyReverse" or "reverse"], i, t_len, fs.id)
				
				if animationStyle.overlap and lastFS and isOverlapping(lastFS, fs) then
					local done = false
					local minimum = percent
					local maximum = 1
					local current = (percent + maximum) / 2
					while maximum - minimum > 0.01 do
						animationStyle.func(fs, scrollArea, current, fs.sticky, i, t_len, fs.id)
						if isOverlapping(lastFS, fs) then
							minimum = current
						else
							maximum = current
						end
						current = (maximum + minimum) / 2
					end
					current = current + 0.01
					fs.start = -current * length + now
					animationStyle.func(fs, scrollArea, current, scrollArea[fs.sticky and "stickyReverse" or "reverse"], i, t_len, fs.id)
					for j = i+1, t_len do
						local v = t[j]
						if v.start > fs.start then
							v.start = fs.start
						end
					end
				end
				lastFS = fs
			end
			if not t[1] then
				u[animationStyle] = del(t)
			end
		end
		if not next(u) then
			wildFontStrings[scrollArea] = del(u)
		end
	end
	if not next(wildFontStrings) then
		wildFontStrings = del(wildFontStrings)
		self:CancelScheduledEvent("Parrot_OnUpdate")
	end
end
